'use strict';

const fs = require('fs');
const path = require('path');
const webpack = require('webpack');
const resolve = require('resolve');
const PnpWebpackPlugin = require('pnp-webpack-plugin');
const HtmlWebpackPlugin = require('html-webpack-plugin');
const CaseSensitivePathsPlugin = require('case-sensitive-paths-webpack-plugin');
const InlineChunkHtmlPlugin = require('react-dev-utils/InlineChunkHtmlPlugin');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const ManifestPlugin = require('webpack-manifest-plugin');
const InterpolateHtmlPlugin = require('react-dev-utils/InterpolateHtmlPlugin');
const WorkboxWebpackPlugin = require('workbox-webpack-plugin');
const getCSSModuleLocalIdent = require('react-dev-utils/getCSSModuleLocalIdent');
const paths = require('./paths');
const modules = require('./modules');
const getClientEnvironment = require('./env');
const ModuleNotFoundPlugin = require('react-dev-utils/ModuleNotFoundPlugin');
const ForkTsCheckerWebpackPlugin = require('react-dev-utils/ForkTsCheckerWebpackPlugin');
const typescriptFormatter = require('react-dev-utils/typescriptFormatter');

const postcssNormalize = require('postcss-normalize');

const appPackageJson = require(paths.appPackageJson);

const shouldUseSourceMap = process.env.GENERATE_SOURCEMAP !== 'false';

const shouldInlineRuntimeChunk = false;

const isExtendingEslintConfig = process.env.EXTEND_ESLINT === 'true';

const imageInlineSizeLimit = parseInt(process.env.IMAGE_INLINE_SIZE_LIMIT || '10000');

const useTypeScript = fs.existsSync(paths.appTsConfig);

const cssRegex = /\.css$/;
const cssModuleRegex = /\.module\.css$/;
const sassRegex = /\.(scss|sass)$/;
const sassModuleRegex = /\.module\.(scss|sass)$/;

module.exports = function(webpackEnv) {
	const isEnvDevelopment = webpackEnv === 'development';
	const isEnvProduction = webpackEnv === 'production';

	const isEnvProductionProfile = isEnvProduction && process.argv.includes('--profile');

	const env = getClientEnvironment(paths.publicUrlOrPath.slice(0, -1), paths.heraPath);

	const getStyleLoaders = (cssOptions, preProcessor) => {
		const loaders = [
			isEnvDevelopment && require.resolve('style-loader'),
			isEnvProduction && {
				loader: MiniCssExtractPlugin.loader,
				options: paths.publicUrlOrPath.startsWith('.') ? { publicPath: '../../' } : {},
			},
			{
				loader: require.resolve('css-loader'),
				options: cssOptions,
			},
			{
				loader: require.resolve('postcss-loader'),
				options: {
					ident: 'postcss',
					plugins: () => [
						require('postcss-flexbugs-fixes'),
						require('postcss-preset-env')({
							autoprefixer: {
								flexbox: 'no-2009',
							},
							stage: 3,
						}),
						postcssNormalize(),
					],
					sourceMap: isEnvProduction && shouldUseSourceMap,
				},
			},
		].filter(Boolean);
		if (preProcessor) {
			loaders.push(
				{
					loader: require.resolve('resolve-url-loader'),
					options: {
						sourceMap: isEnvProduction && shouldUseSourceMap,
					},
				},
				{
					loader: require.resolve(preProcessor),
					options: {
						sourceMap: true,
					},
				},
			);
		}
		return loaders;
	};

	return {
		mode: isEnvProduction ? 'production' : isEnvDevelopment && 'development',
		bail: isEnvProduction,
		devtool: isEnvProduction ? false : isEnvDevelopment && 'cheap-module-source-map',
		entry: [isEnvDevelopment && require.resolve('react-dev-utils/webpackHotDevClient'), paths.appIndexJs].filter(
			Boolean,
		),
		output: {
			path: isEnvProduction ? paths.appBuild : undefined,
			pathinfo: isEnvDevelopment,
			filename: isEnvProduction ? 'static/js/[name].js' : isEnvDevelopment && 'static/js/bundle.js',
			futureEmitAssets: true,
			chunkFilename: isEnvProduction
				? 'static/js/[name].[contenthash:8].chunk.js'
				: isEnvDevelopment && 'static/js/[name].chunk.js',
			publicPath: paths.publicUrlOrPath,
			devtoolModuleFilenameTemplate: isEnvProduction
				? info => path.relative(paths.appSrc, info.absoluteResourcePath).replace(/\\/g, '/')
				: isEnvDevelopment && (info => path.resolve(info.absoluteResourcePath).replace(/\\/g, '/')),
			jsonpFunction: `webpackJsonp${appPackageJson.name}`,
			globalObject: 'this',
		},
		optimization: {
			minimize: isEnvProduction,
		},
		resolve: {
			modules: ['node_modules', paths.appNodeModules].concat(modules.additionalModulePaths || []),
			extensions: paths.moduleFileExtensions.map(ext => `.${ext}`).filter(ext => useTypeScript || !ext.includes('ts')),
			alias: {
				...(modules.webpackAliases || {}),
				Hera: path.resolve(__dirname, 'hera/core.js'),
			},
			plugins: [PnpWebpackPlugin],
		},
		resolveLoader: {
			plugins: [PnpWebpackPlugin.moduleLoader(module)],
		},
		module: {
			strictExportPresence: true,
			rules: [
				{ parser: { requireEnsure: false } },
				{
					oneOf: [
						{
							test: [/\.bmp$/, /\.gif$/, /\.jpe?g$/, /\.png$/],
							loader: require.resolve('url-loader'),
							options: {
								limit: imageInlineSizeLimit,
								name: 'static/media/[name].[ext]',
							},
						},
						{
							test: /\.(js|mjs|jsx|ts|tsx)$/,
							include: paths.appSrc,
							loader: require.resolve('babel-loader'),
							options: {
								customize: require.resolve('babel-preset-react-app/webpack-overrides'),

								plugins: [
									[
										require.resolve('babel-plugin-named-asset-import'),
										{
											loaderMap: {
												svg: {
													ReactComponent: '@svgr/webpack?-svgo,+titleProp,+ref![path]',
												},
											},
										},
									],
								],
								cacheDirectory: true,
								cacheCompression: false,
								compact: isEnvProduction,
							},
						},
						{
							test: /\.(js|mjs)$/,
							exclude: /@babel(?:\/|\\{1,2})runtime/,
							loader: require.resolve('babel-loader'),
							options: {
								babelrc: false,
								configFile: false,
								compact: false,
								presets: [[require.resolve('babel-preset-react-app/dependencies'), { helpers: true }]],
								cacheDirectory: true,
								cacheCompression: false,
								sourceMaps: shouldUseSourceMap,
								inputSourceMap: shouldUseSourceMap,
							},
						},
						{
							test: cssRegex,
							exclude: cssModuleRegex,
							use: getStyleLoaders({
								importLoaders: 1,
								sourceMap: isEnvProduction && shouldUseSourceMap,
							}),
							sideEffects: true,
						},
						{
							test: cssModuleRegex,
							use: getStyleLoaders({
								importLoaders: 1,
								sourceMap: isEnvProduction && shouldUseSourceMap,
								modules: {
									getLocalIdent: getCSSModuleLocalIdent,
								},
							}),
						},
						{
							test: sassRegex,
							exclude: sassModuleRegex,
							use: getStyleLoaders(
								{
									importLoaders: 3,
									sourceMap: isEnvProduction && shouldUseSourceMap,
								},
								'sass-loader',
							),
							sideEffects: true,
						},
						{
							test: sassModuleRegex,
							use: getStyleLoaders(
								{
									importLoaders: 3,
									sourceMap: isEnvProduction && shouldUseSourceMap,
									modules: {
										getLocalIdent: getCSSModuleLocalIdent,
									},
								},
								'sass-loader',
							),
						},
						{
							test: /\.hera$/i,
							use: 'raw-loader',
						},
						{
							loader: require.resolve('file-loader'),
							exclude: [/\.(js|mjs|jsx|ts|tsx)$/, /\.html$/, /\.json$/],
							options: {
								name: 'static/media/[name].[ext]',
							},
						},
					],
				},
			],
		},
		plugins: [
			new HtmlWebpackPlugin(
				Object.assign(
					{},
					{
						inject: true,
						template: paths.appHtml,
					},
					isEnvProduction
						? {
								minify: {
									removeComments: true,
									collapseWhitespace: true,
									removeRedundantAttributes: true,
									useShortDoctype: true,
									removeEmptyAttributes: true,
									removeStyleLinkTypeAttributes: true,
									keepClosingSlash: true,
									minifyJS: true,
									minifyCSS: true,
									minifyURLs: true,
								},
						  }
						: undefined,
				),
			),
			isEnvProduction && shouldInlineRuntimeChunk && new InlineChunkHtmlPlugin(HtmlWebpackPlugin, [/runtime-.+[.]js/]),
			new InterpolateHtmlPlugin(HtmlWebpackPlugin, env.raw),
			new ModuleNotFoundPlugin(paths.appPath),
			new webpack.DefinePlugin(env.stringified),
			isEnvDevelopment && new webpack.HotModuleReplacementPlugin(),
			isEnvDevelopment && new CaseSensitivePathsPlugin(),
			// isEnvDevelopment && new WatchMissingNodeModulesPlugin(paths.appNodeModules),
			isEnvProduction &&
				new MiniCssExtractPlugin({
					filename: 'static/css/[name].css',
				}),
			new ManifestPlugin({
				fileName: 'asset-manifest.json',
				publicPath: paths.publicUrlOrPath,
				seed: {
					'codeMirror.js': '/codeMirror/hera.js',
				},
				generate: (seed, files, entrypoints) => {
					const manifestFiles = files.reduce((manifest, file) => {
						manifest[file.name] = file.path;
						return manifest;
					}, seed);
					const entrypointFiles = entrypoints.main.filter(fileName => !fileName.endsWith('.map'));

					return {
						files: manifestFiles,
						entrypoints: entrypointFiles,
					};
				},
			}),
			new webpack.IgnorePlugin(/^\.\/locale$/, /moment$/),
			isEnvProduction &&
				new WorkboxWebpackPlugin.GenerateSW({
					clientsClaim: true,
					exclude: [/\.map$/, /asset-manifest\.json$/],
					importWorkboxFrom: 'cdn',
					navigateFallback: paths.publicUrlOrPath + 'index.html',
					navigateFallbackBlacklist: [
						// Exclude URLs starting with /_, as they're likely an API call
						new RegExp('^/_'),
						// Exclude any URLs whose last part seems to be a file extension
						// as they're likely a resource and not a SPA route.
						// URLs containing a "?" character won't be blacklisted as they're likely
						// a route with query params (e.g. auth callbacks).
						new RegExp('/[^/?]+\\.[^/]+$'),
					],
				}),
			useTypeScript &&
				new ForkTsCheckerWebpackPlugin({
					typescript: resolve.sync('typescript', {
						basedir: paths.appNodeModules,
					}),
					async: isEnvDevelopment,
					useTypescriptIncrementalApi: true,
					checkSyntacticErrors: true,
					resolveModuleNameModule: process.versions.pnp ? `${__dirname}/pnpTs.js` : undefined,
					resolveTypeReferenceDirectiveModule: process.versions.pnp ? `${__dirname}/pnpTs.js` : undefined,
					tsconfig: paths.appTsConfig,
					reportFiles: [
						'**',
						'!**/__tests__/**',
						'!**/?(*.)(spec|test).*',
						'!**/src/setupProxy.*',
						'!**/src/setupTests.*',
					],
					silent: true,
					formatter: isEnvProduction ? typescriptFormatter : undefined,
				}),
		].filter(Boolean),
		node: {
			module: 'empty',
			dgram: 'empty',
			dns: 'mock',
			fs: 'empty',
			http2: 'empty',
			net: 'empty',
			tls: 'empty',
			child_process: 'empty',
		},
		performance: false,
	};
};
